module Moneda where

import Array

-- VORAZ

-- devuelve solo el numero de monedas
voraz :: [Int] -> Int -> Int
voraz ds 0 = 0
voraz ds m = cociente + voraz ds resto
           where dmax = maximum (filter (<=m) ds)
                 cociente = m `div` dmax
                 resto = m `mod` dmax

-- devuelve la lista de monedas
voraz_detallado :: [Int] -> Int -> [(Int,Int)]
voraz_detallado ds 0 = []
voraz_detallado ds m = (dmax,cociente) : voraz_detallado ds resto
                     where dmax = maximum (filter (<=m) ds)
                           cociente = m `div` dmax
                           resto = m `mod` dmax

-- BACKTRACKING

-- infinito es un número suficientemente grande
infinito = 2^30 :: Int

-- las siguientes funciones solo devuelven el número óptimo de monedas

-- corresponde a primera version de las filminas, con indices i y j 
backtracking :: [Int] -> Int -> Int
backtracking ds k = m (length ds) k
                  where m i 0 = 0
                        m 0 j = infinito
                        m i j | j < ds!!(i-1) = m (i-1) j
                              | otherwise = m (i-1) j `min` (1 + m i (j - ds!!(i-1)))

-- una variante que encaja más con el estilo de Haskell,
-- reemplazamos el indice i, que se usaba para indexar la lista de denominaciones
-- por la propia lista de denominaciones 
backtracking1 :: [Int] -> Int -> Int
backtracking1 ds k = m ds k
                   where m ds 0 = 0
                         m [] j = infinito
                         m (d:ds) j | j < d = m ds j
                                    | otherwise = m ds j `min` (1 + m (d:ds) (j-d))

-- corresponde a la segunda version de las filminas
-- como backtracking1 usa la lista de denominaciones en lugar del indice i
backtracking2 :: [Int] -> Int -> Int
backtracking2 ds k = m ds k
                   where m ds 0 = 0
                         m [] j = infinito
                         m (d:ds) j = minimum (map (\q -> q + m ds (j-q*d)) [0..j `div` d])

-- corresponde a la tercera (y ultima) version de las filminas
backtracking3 :: [Int] -> Int -> Int
backtracking3 ds n = m (length ds-1) n
                   where m i 0 = 0
                         m i j = minimum (map (\i' -> if (ds!!i') > j
                                                      then infinito
                                                      else (1 + m i' (j-(ds!!i')))) [0..i])

-- a partir de acá, se devuelve la lista de monedas

-- esta lista es para usar cuando no hay solucion
sin_solucion :: [Int]
sin_solucion = [-1]

-- para comparar dos soluciones, cuál es más corta
-- la que representa que no hay solución tiene que perder en la comparación
mas_corta :: [Int] -> [Int] -> [Int]
mas_corta s1 s2 | any ((-1)==) s1 = s2
                | any ((-1)==) s2 = s1
                | otherwise = if length s1 <= length s2 then s1 else s2

-- dos técnicas:
-- 1) acarreando un parámetro con la solución parcial
backtracking1_detallado :: [Int] -> Int -> [Int]
backtracking1_detallado ds n = m [] ds n
                             where m s ds 0 = s
                                   m s [] j = sin_solucion
                                   m s (d:ds) j | j < d = m s ds j
                                                | otherwise = let s1 = m s ds j
                                                                  s2 = m (d:s) (d:ds) (j-d)
                                                              in mas_corta s1 s2
-- 2) sin acarrearlo
backtracking_detallado :: [Int] -> Int -> [Int]
backtracking_detallado ds n = m ds n
                            where m ds 0 = []
                                  m [] j = sin_solucion
                                  m (d:ds) j | j < d = m ds j
                                             | otherwise = let s1 = m ds j
                                                               s2 = d : m (d:ds) (j-d)
                                                           in mas_corta s1 s2

-- PROGRAMACION DINAMICA

-- devuelve solo el numero de monedas
dinamica :: [Int] -> Int -> Int
dinamica ds n = tabla!(k,n)
              where k = length ds
                    tabla = array ((0,0),(k,n)) [((i,j),m i j) | i <- [0..k], j <- [0..n]]
                    m i 0 = 0
                    m 0 j = infinito
                    m i j | j < (ds!!(i-1)) = tabla!(i-1,j)
                          | otherwise = tabla!(i-1,j) `min` (1+tabla!(i,j-(ds!!(i-1))))

-- devuelve la lista de monedas
dinamica_detallado :: [Int] -> Int -> [Int]
dinamica_detallado ds n = tabla!(k,n)
                        where k = length ds
                              tabla = array ((0,0),(k,n)) [((i,j),m i j) | i <- [0..k], j <- [0..n]]
                              m i 0 = []
                              m 0 j = sin_solucion
                              m i j | j < (ds!!(i-1)) = tabla!(i-1,j)
                                    | otherwise = let s1 = tabla!(i-1,j)
                                                      s2 = ds!!(i-1) : tabla!(i,j-(ds!!(i-1)))
                                                  in mas_corta s1 s2

-- devuelve la lista de monedas, con la técnica del teórico
dinamica_recupero  :: [Int] -> Int -> [Int]
dinamica_recupero ds n = recupero k n
                       where k = length ds
                             tabla = array ((0,0),(k,n)) [((i,j),m i j) | i <- [0..k], j <- [0..n]]
                             m i 0 = 0
                             m 0 j = infinito
                             m i j | j < (ds!!(i-1)) = tabla!(i-1,j)
                                   | otherwise = tabla!(i-1,j) `min` (1+tabla!(i,j-(ds!!(i-1))))
                             recupero i 0 = []
                             recupero i j | tabla!(i,j) == tabla!(i-1,j) = recupero (i-1) j
                                          | otherwise = ds!!(i-1) : recupero i (j-ds!!(i-1))


